package provide mmucl 1.01

##########################################################################
# mmucl.tcl - base mmucl library
#
# Copyright (C) 1997-1998 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##########################################################################

# This package provides the basic procedures for making a mud client.
# User commands are executed in a seperate interpreter, mmucl.
# All commands in the mmucl namespace in the form MCname are made
# available to the user as aliases in the mmucl interpreter.
# The package should be used by first making the namespace mmucl
# and in it defining the callback mmucl::display and any
# additional user commands. Then mmucl::init should be called.

global mmucl
array set mmucl {
    reconnect  0
    host       ""
    port       ""
    char       ""
    no_action  0
    no_sub     0
}

# The user can add to or modify existing elements using "config".
# User commands should use this array to allow configuration.

#  reconnect  - if non-zero, reconnect after timeout or disconnect
#  host       - mud host
#  port       - mud port
#  char       - current character loaded by charload
#  no_action  - whether actions are checked
#  no_sub     - whether subs are checked

# array mapping text attribs to ansi codes
array set ansi {
    reset 0 bold 1 underline 4 blink 5 reverse 7

    black 30 red 31 green 32 yellow 33 brown 33 blue 34
    magenta 35 cyan 36 grey 37 gray 37

    bg_black 40 bg_red 41 bg_green 42 bg_yellow 43 bg_brown 43
    bg_blue 44 bg_magenta 45 bg_cyan 46 bg_grey 47 bg_gray 47
}

namespace eval mmucl {
    namespace export init
}

# mmucl::init --
#
#  Initialize mmucl. Make an interpreter to execute usr commands. Load
#  a usr configuration file into the interpreter and make commands
#  prefixed by MC in the mmucl namespace available to the user.
#  Init checks argv for command line arguments.
#    -safe makes the user interpreter a safe interpreter.
#    -exec arg ?arg ...? executes the given tcl script in the mmucl
#         interpreter.

proc mmucl::init {argv} {
    variable Priv
    variable Char
    global env tcl_library auto_path 

    set i [lsearch $argv -safe]
    if {$i != -1} {
	set argv [lreplace $argv $i $i]
	interp create -safe mmucl
	mmucl eval [list set tcl_library $tcl_library]
	mmucl eval [list set auto_path $auto_path]
	interp invokehidden mmucl source [file join $tcl_library history.tcl]
    } else {
	interp create mmucl
    }
    
    foreach cmd [info commands ::mmucl::MC*] {
    	interp alias mmucl [string range $cmd 11 end] {} $cmd
    }

    set Priv(connect) 0
    set Priv(rc_dir) [file join $env(HOME) .mmucl]
    
    # Make rc_dir if it doesn't exist.
    if {![file exists $Priv(rc_dir)]} {
	file mkdir $Priv(rc_dir)
    }
    
    set rc [file join $Priv(rc_dir) mmucl.rc]
    if {[file exists $rc]} {
	loadFile $rc
    }

    # Load defined characters.
    if {[file exists [file join $Priv(rc_dir) charlist]]} {
	source [file join $Priv(rc_dir) charlist]
    }
    
    MCecho "Mmucl version 1.0b6\n" {underline}

    set i [lsearch $argv -exec]
    if {$i != -1} {
	catch  {mmucl eval [lrange $argv [incr i] end]}
    }
}

# display has to be defined by the app.
# Use the below as a guide.

# mmucl::display --
#
#   Display a string sent from the mud. Used by read_mud.

#proc mmucl::display {str} 
#    puts -nonewline $str
#    flush stdout
#}

# mmucl::MCecho --
#
#   Display the given string.

proc mmucl::MCecho {str {attribs {}}} {
    global mmucl
    
    display [toansi $str $attribs]\n
}

# mmucl::toansi --

# 
#   Give str ansi attributes and return it.   

proc mmucl::toansi {str attribs} {
    global ansi

    set new \033\[0
    foreach attrib [string tolower $attribs] {
	if {[info exists ansi($attrib)]} {
	    append new \;$ansi($attrib)
	}
    }
    append new m$str\033\[0\;37\;40m
    return $new
}


# mmucl::MCmatch --
#
#   Do substitution on the text. Then check the text to see if it
#   matches any actions. If it does execute the action scripts,
#   substituting %n in the commands for the nth parenthesized 
#   subexpression in the action pattern. %n can be escaped with \ 

proc mmucl::MCmatch {str} {
    global mmucl
    variable Action
    variable Sub

    if {![string compare $mmucl(no_sub) 0]} { 
	foreach {patt replace} [array get Sub] {
	    regsub -all -- $patt $str [join $replace] str
	}
    }
    
    if {![string compare $mmucl(no_action) 0]} {
    foreach {pat script} [array get Action] {
	set substr $str
	while {[regexp -nocase -indices -- $pat $substr 0 1 2 3 4 5 6 7 8 9]} {
	    regsub -all {[\[$]} [join $script] {\\&} script
	    regsub -all {([^\\]?)%([0-9])} $script \
		{\1[eval string range {$substr} $\2]} script
	    regsub -all {\\(%[0-9])} $script {\1} script
	    if {[catch {mmucl eval [subst $script]} error]} {
		MCecho "** Action {$pat} error: $error **"
	    }
	    set substr [string range $substr [expr {[lindex $0 1]+1}] end]
	}
    }
    }
    return $str
}

# An alternate implementation of match. Sub expression matches
# in this version are \n, like the regsub subspec.
# It looks nicer but it is slower and doesn't quite work. :)

# proc mmucl::MCmatch {str} {
#     variable Action
#     variable Sub
    
#     foreach {patt script} [array get Action] {
# 	if {[regsub -all -nocase -- $patt $str \n[join $script]\n\# script]} {
# 	    if {[catch {mmucl eval \#$script} error]} {
# 	       MCecho "** Action {$patt} error: $error **"
# 	   }
# 	}
#     }
    
#     foreach {patt replace} [array get Sub] {
# 	regsub -all -- $patt $str [join $replace] str
#     }
#     return $str
# }


# mmucl::MCparse --
#
#   Check input to see what should be done with it.
#     \string means send string to the mud as is minus the "\"
#     #string means string is a tcl script.

#     Input is considered to consist of commands seperated
#     by ";". Each command is checked to see if it matches an
#     alias and if so the alias is executed.

#   Have to do a special case for "" (return) because
#   the split eats it up.
 
proc mmucl::MCparse {str} {
    variable Alias

    set fchar [string index $str 0]
    if {$fchar=="\\"} {
	MCmudputs [string range $str 1 end]
    } elseif {$fchar=="#"} {
	catch {set ret [mmucl eval [string range $str 1 end]]} ret
	MCecho $ret
    } elseif {$str==""} {
	MCmudputs ""
    } else {
	foreach cmd [split $str \;] {
	    regexp {^([^ ]*) *(.*)$} $cmd x alias args
	    if {[info exists Alias($alias)]} { 
		MCexec_alias $alias $args
	    } else {
		MCmudputs $cmd
	    }
	}
    }
}

# mmucl::MCconnect --
#
#   Connect to a mud. If given a third argument, it's written
#   to the mud on connect.
#   A connection will time out after 20 seconds.
#   If the var mmucl(reconnect) is non-zero the proc will attempt
#   to reconnect in 2 seconds.
#   Is the telnet protocol worth dealing with?

proc mmucl::MCconnect {host port {login ""}} {
    global mmucl
    variable Priv

    if {$Priv(connect)} {
	MCecho "** Already connected. **"
	return
    }

    MCecho "** Connecting to $host at port $port... **"
    
    # Do the close and after cancel to make sure
    # we're only attempting one connect at a time.
    catch {close $Priv(sockid); after cancel $Priv(timeout_id)}
    
    # Set up a timeout.
    set Priv(timeout_id) [after 20000 {
	if {!$mmucl::Priv(connect)} {
	    catch {close $mmucl::Priv(sockid)}
	    mmucl::MCecho "** Timed out. **"
	    if {[string compare $mmucl(reconnect) 0]} {
		after 2000 mmucl::MCreconnect
	    }
	}}]
    
    if {[catch {set Priv(sockid) [socket -async $host $port]} error]} {
	after cancel $Priv(timeout_id)
	MCecho "** Error: $error **"
    } else {
	fconfigure $Priv(sockid) -blocking 0 -translation binary
	set mmucl(host) $host
	set mmucl(port) $port
	
	# Fileevent only gets called once a connect occurs.
	# It sets a variable the after commands check to see if we
	# connect in a specific period of time and times out if we don't.
	
	fileevent $Priv(sockid) readable "
	    after cancel $mmucl::Priv(timeout_id)
	    unset mmucl::Priv(timeout_id)
	    set mmucl::Priv(connect) 1
	    fileevent $mmucl::Priv(sockid) readable mmucl::read_mud
	    mmucl::read_mud
	    if {\$mmucl::Priv(connect)} {mmucl::MCwritemud {$login}}
           "
    }
}

# mmucl::MCdisconnect --
#
#   Disconnect from a mud. 

proc mmucl::MCdisconnect {} {
    global mmucl
    variable Priv

    if {!$Priv(connect)} {
	MCecho "** Not connected **"
    } else {
	catch {close $Priv(sockid)}
	set Priv(connect) 0
	MCecho "** Connection closed. **"
	if {[string compare $mmucl(reconnect) 0]} {
	    after 2000 mmucl::MCreconnect
	}
    }
}


# mmucl::MCreconnect --
#
#   Try to resume previous connect using mmucl(host) and mmucl(port) and
#   mmucl(char) to get a login string..

proc mmucl::MCreconnect {} {
    global mmucl
    variable Char
  
    if {[info exists mmucl(host)] && [info exists mmucl(port)]} {
	set login ""
	catch {set login [lindex $Char($mmucl(char)) 3]}
	MCconnect $mmucl(host) $mmucl(port) $login
    } else {
	MCecho "** No previous connection **"
    }
}

# mmucl::MCconfig --
#
#   Give the user the ability to look at and modify the
#   global array mmucl. Elements can be queried or modified,
#   but not added.

proc mmucl::MCconfig {args} {
    global mmucl
    
    set n [llength $args]
    if {$n > 2} {
	return -code error "wrong # args: config ?option? ?value?"
    }

    if {$n == 0} {
	return [array names mmucl]
    }

    set option [lindex $args 0]

    if {$n == 2} {
	return [set mmucl($option) [lindex $args 1]]
    }

    if {![info exists mmucl($option)]} {
	return -code error "\"$option\" isn't a valid option"
    } else {
	return $mmucl($option)
    }
}

# mmucl::MCwritemud
#
#   Break up str into commands seperated by ; and
#   write them to the mud.

proc mmucl::MCwritemud {str} {
    variable Priv
    
    if {!$Priv(connect)} {
	MCecho "** Not connected. **"
    } else {
	foreach cmd [split $str \;] {
	    puts $Priv(sockid) $cmd
	}
	flush $Priv(sockid)
    }
}

# mmucl::MCmudput --
#
#   Write str to the mud

proc mmucl::MCmudputs {str} {
    variable Priv
    
    if {!$Priv(connect)} {
	MCecho "** Not connected. **"
    } else {
	puts $Priv(sockid) $str
	flush $Priv(sockid)
    }
}

# mmucl::MCalias
#
#   Add or display aliases

proc mmucl::MCalias {{name *} {script ""}} {
    variable Alias
    
    if {$script==""} {
	set result [array names Alias $name]
	if {$result !=""} {
	    foreach match [lsort $result] {
		MCecho "** Alias $match: $Alias($match) **"
	    }
	} else {
	    MCecho "** No alias matching: $name **"
	}
    } else {
	set Alias($name) [list $script]
	MCecho "** Alias \"$name\" set to {$script} **"
    }
}

# mmucl::MCunalias --
#
#   Delete an alias

proc mmucl::MCunalias {name} {
    variable Alias
    
    set result [array names Alias $name]
    if {$result !=""} {
	foreach match $result {
	    unset Alias($match)
	    MCecho "** \"$match\" is no longer an alias. **"
	}
    } else {
	MCecho "** No alias matching: $name. **"
    }
}

# mmucl::MCaction --
#
#   Add or display actions

proc mmucl::MCaction {{pattern *} {script ""}} {
    variable Action

    if {$script==""} {
	set result [array names Action $pattern]
	if {$result !=""} {
	    foreach match [lsort $result] {
		MCecho "** Action \{$match\}: $Action($match) **"
	    }
	} else {
	    MCecho "** No action matching: $pattern **"
	}
    } else {
        # Check to see if the pattern is a valid regexp.
        if {[catch {regexp $pattern ""} error]} {
            MCecho "** $error"
	    return
	}
        set Action($pattern) [list $script]
	MCecho "** Action {$pattern} triggers {$script} **"
    }
}

# mmucl::MCunaction --
#
#   delete an action

proc mmucl::MCunaction {pattern} {
    variable Action

    # make sure pattern doesn't blow up in [array names...]
    regsub -all {[\[]} $pattern {\\&} pattern
    
    set result [array names Action $pattern]
    if {$result !=""} {
	foreach match $result {
	    unset Action($match)
	    MCecho "** {$match} is no longer an action. **"
	}
    } else {
	MCecho "** No action matching: $pattern. **"
    }
}

# mmucl::MCsub --
#
#   Add or display substitutions
#   Ugly because have to allow replacement of string with ""

proc mmucl::MCsub {{pattern *} args} {
    variable Sub

    set n [llength $args]
    if {$n > 1} {
	return -code error "called \"sub\" with too many arguments"
    }

    if {$n < 1} {
	set result [array names Sub $pattern]
	if {$result !=""} {
	    foreach match [lsort $result] {
		MCecho "** Substitution {$match}: $Sub($match) **"
	    }
	} else {
	    MCecho "** No substitutions matching: $pattern **"
	}
    } else {
        # Check to see if the pattern is a valid regexp.
        if {[catch {regexp $pattern ""} error]} {
            MCecho "** $error"
	    return
	}
        set Sub($pattern) $args
	MCecho "** Substitution \"$pattern\" for \"[lindex $args 0]\" **"
    }
}

# mmucl::MChighlight --
#
#   Add ansi attributes the match of a pattern.

proc mmucl::MChighlight {pattern attribs} {
    MCsub $pattern [toansi & $attribs]
}

# mmucl::MCunsub --
#
#   delete a substitution

proc mmucl::MCunsub {pattern} {
    variable Sub

    # make sure pattern doesn't blow up in [array names...]
    regsub -all {[\[]} $pattern {\\&} pattern
    
    set result [array names Sub $pattern]
    if {$result !=""} {
	foreach match $result {
	    unset Sub($match)
	    MCecho "** \"$match\" is no longer a substitution. **"
	}
    } else {
	MCecho "** No substitutions matching: $pattern. **"
    }
}

# mmucl::MCexec_alias --
#
#   Substitute %n for nth arg to the alias in the alias command
#   and execute it.

proc mmucl::MCexec_alias {name arg_list} {
    variable Alias

    if {![info exists Alias($name)]} {
	return -code error "Alias $name doesn't exist"
    }

    if {[catch {set 0 [join $arg_list]} error]} {
	MCecho "Error in alias arguments: $error"
	return
    }

    # Make sure we set the variables to some value
    lappend arg_list {}
    foreach {1 2 3 4 5 6 7 8 9} $arg_list {}

    regsub -all {[\[$]} [join $Alias($name)] {\\&} script
    regsub -all {([^\\]?)%([0-9])} $script {\1$\2} script
    regsub -all {\\(%[0-9])} $script {\1} script

    if {[catch {mmucl eval [subst $script]} error]} {
	MCecho "** Alias \"$name\" error: $error **"
    }
}

# mmucl::MCcharadd --
#
#   Defined a character to be loaded with charload.

proc mmucl::MCcharadd {char mud host port {login ""}} {
    variable Priv
    variable Char

    set Char($char) [list $mud $host $port $login]
    set init_file [file join $Priv(rc_dir) $char.$mud]

    # Don't overwrite an init file if it's already there
    if {![file exists $init_file]} {
	if {[catch {set fd [open $init_file w]} error]} {
	    MCecho "** Error: $error **"
	    return
	}
	close $fd
    }

    MCecho "** Created character $char. **"
    MCecho "** Init file: $init_file **"   
}

# mmucl::MCcharload --
#
#   Connect to defined char's mud and load his init file.

proc mmucl::MCcharload {{name ""}} {
    global mmucl
    variable Char
    variable Priv

    if {![info exists Char($name)]} {
	MCecho "** No character \"$name\" defined. **"
	return
    }
    foreach {mud host port login} $Char($name) {}

    set init_file [file join $Priv(rc_dir) $name.$mud]
    if {[file exists $init_file]} {
	loadFile $init_file
    } else {
	MCecho "** Init file $init_file not found **"
    }
    
    MCconnect $host $port $login
    set mmucl(char) $name
}

# mmucl::MCchardelete --
#
#   Delete a character

proc mmucl::MCchardelete {name} {
    variable Priv
    variable Char

    if {![info exists Char($name)]} {
	MCecho "** No defined character $name. **"
	return
    }

    set mud [lindex $Char($name) 0]
    unset Char($name)

    # Should we actually delete the init file?
    catch {file delete [file join $Priv(rc_dir) $name.$mud]}
   
    MCecho "** Character $name deleted. **"
}

# mmucl::MCsave --
#
#   Try to save current state to a file.
#   Worth doing procs and variables? Make args available
#   selecting what to save?

proc mmucl::MCsave {filename} {
    variable Alias
    variable Action
    variable Sub

    if {[catch {set fd [open $filename w]} error]} {
	MCecho "** Error: $error **"
	return
    }
    foreach alias [array names Alias] {
	puts $fd "alias {$alias} $Alias($alias)"
    }
    foreach action [array names Action] {
	puts $fd "action {$action} $Action($action)"
    }
    foreach sub [array names Sub] {
	puts $fd "sub {$sub} $Sub($sub)"
    }
    flush $fd
    close $fd
    MCecho "** Mmucl state saved in $filename. **"
}

# mmucl::MCtextin --
#
#   Read a file directly into the mud.

proc mmucl::MCtextin {filename} {
    if {[catch {set fileid [open $filename r]} error]} {
	MCecho "** $error **"
    } else {
	MCmudputs [read $fileid]
	close $fileid
    }
}


# mmucl::MCexit --
#
#   Exit the app and write charlist.

proc mmucl::MCexit {} {
    variable Priv
    variable Char
    
    # Write charlist before we exit.
    set fileid [open [file join $Priv(rc_dir) charlist] w]
    puts $fileid [list array set Char [array get Char]] 
    flush $fileid
    close $fileid

    exit
}

# mmucl::read_mud  --
#
#   Get output from mud, check it for actions with match, and display it
#   with display.

proc mmucl::read_mud {} {
    variable Priv
 
    if {[catch {set mudout [read $Priv(sockid)]} err] || [eof $Priv(sockid)]} {
	if {$err != ""} {
	    MCecho "** Error: $err **"
	}
	MCdisconnect
    } else {
	display [MCmatch $mudout]
    }
}

# mmucl::loadFile --
#
#   Load a file into the user interpreter, dealing with a possible
#   safe interp.

proc mmucl::loadFile {file} {
    global errorInfo

    if {![mmucl issafe]} {
	if {[catch {mmucl eval [list source $file]} error]} {
	    MCecho "Error: $error\nStack trace:\n$errorInfo"
	}
    } else {
	if {[catch {interp invokehidden mmucl -global source $file} error]} {
	    MCecho "Error: $error\nStack trace:\n$errorInfo"
	}
    }
}
